/* Copyright 2008 Daniel Vokó <vokoda@gmail.com>
 * 
 * This file is part of Sagittarius.
 *
 * Sagittarius is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sagittarius is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Sagittarius.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using Sagittarius.Exiftool;
using Sagittarius.Metadata;
using System.Collections;
using System.Threading;
using Gdk;

namespace Sagittarius
{
	public class PhotoInfoFactory
	{

		public static ArrayList threadsToStop;

		private Queue photoQueue;
		private Queue photoDone;
		private Thread factoryWorker;
		private FactorySemaphore factorySemaphore;
		private Gtk.ThreadNotify threadNotify;
		private Config config;
		private CacheManager cacheManager;

		static PhotoInfoFactory ()
		{
			threadsToStop = new ArrayList ();
		}

		public static void KillThreads ()
		{
			foreach (Thread t in threadsToStop) 
			{
				t.Abort ();
			}
		}

		public PhotoInfoFactory(Config config,CacheManager cacheManager)
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory constructor");
			this.config=config;
			this.cacheManager=cacheManager;
			photoQueue = new Queue ();
			photoDone = new Queue ();
			factorySemaphore = new FactorySemaphore ();

			ThreadStart tStart = new ThreadStart (FactoryThread);
			factoryWorker = new Thread (tStart);
			factoryWorker.Start ();

			threadsToStop.Add (factoryWorker);

			threadNotify = new Gtk.ThreadNotify (new Gtk.ReadyEvent (ThumbDataFinished));
			//Console.WriteLine("Leaving Sagittarius.ThumbDataFactory constructor");
		}

		public void BeginProcessPhoto (string filename)
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory.BeginProcessPhoto for file: "+filename);
			// a stack nem thread_safe, ezért lockolni kell
			lock (photoQueue) 
			{
				// ha a stackben nincs benne az adott thumbnailInfo ojjekt, 
				//if (!photoStack.Contains (pInfo)) 
				//{
				PhotoInfo pInfo = new PhotoInfo(filename,config,cacheManager);
					// akkor belerakjuk
					photoQueue.Enqueue (pInfo);
					// majd értesítjük a sorban álló thread-eket, 
					// hogy a szemafor szabad ???
					factorySemaphore.Up ();
				//}
			}
			//Console.WriteLine("Leaving Sagittarius.ThumbDataFactory.BeginProcessPhoto for file: "+filename);
		}

		private void ThumbDataFinished ()
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory.ProcessFinished");
			lock (photoDone) 
			{
				while (photoDone.Count > 0) 
				{
					PhotoInfo pInfo = (PhotoInfo) photoDone.Dequeue ();
					if (OnPhotoInfoReady != null)
					{
						OnPhotoInfoReady (this, new PhotoInfoReadyEventArgs (pInfo));
					}
				}
			}
			//Console.WriteLine("Leaving Sagittarius.ThumbDataFactory.ProcessFinished");
		}
		
		private void FactoryThread ()
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory.FactoryThread");
			while (true) 
			{
				try 
				{
					factorySemaphore.Down ();
				} 
				catch 
				{}

				PhotoInfo pInfo;

				while (photoQueue.Count > 0) 
				{
					lock (photoQueue) 
					{
						pInfo = (PhotoInfo) photoQueue.Dequeue ();
					}

					ProcessPhoto (pInfo);

					lock (photoDone) 
					{
						photoDone.Enqueue (pInfo);
					}
					threadNotify.WakeupMain ();
				}
			}
		}
		
		private void ProcessPhoto (PhotoInfo pInfo)
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory.ProcessPhoto for file: "+pInfo.Filename);
			CacheInfo cInfo = cacheManager.GetCacheForFile(pInfo.Filename);

			pInfo.IsPixbuf=DetermineType(pInfo.Filename);

			if(cInfo==null)
			{
				//Console.WriteLine("File is not cached, creating...");
				
				if(pInfo.IsPixbuf)
				{
					using (Pixbuf pixbuf = new Pixbuf (pInfo.Filename)) 
					{
						//GenerateThumbAndPreview(pInfo,pixbuf);
						GenerateThumb(pInfo,pixbuf);
					}
				}
				else
				{
					ExiftoolWrapper exiftoolWrapper = new ExiftoolWrapper(config.ExitoolExecutable);
					using(Pixbuf pixbuf = new Pixbuf((System.IO.Stream)exiftoolWrapper.GetBinaryTagFromFile(pInfo.Filename,"preview")))
					{
						GenerateThumb(pInfo,pixbuf);
						//GenerateThumbAndPreview(pInfo,pixbuf);
					}
					exiftoolWrapper = null;
				}
				cacheManager.AddCache(pInfo.Filename,pInfo.Thumbnail);
			}
			else
			{
				if(cInfo.previewFilename!="")
				{
					pInfo.Preview=new Pixbuf(cInfo.previewFilename);
				}
				pInfo.Thumbnail= new Pixbuf(cInfo.thumbFilename);
			}

			pInfo.Metadata = new XmpMetadata(pInfo.Filename,new ExiftoolWrapper(config.ExitoolExecutable));

			//Console.WriteLine("Leaving Sagittarius.ThumbDataFactory.ProcessPhoto for file: "+pInfo.Filename);
		}
		private void GenerateThumb(PhotoInfo pInfo, Pixbuf pixbuf)
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory.GenerateThumb");
			int photoWidth= pixbuf.Width;
			int photoHeight= pixbuf.Height;
			int thumbWidth;
			int thumbHeight;
			
			if(photoWidth>=photoHeight)
			{
				thumbHeight=(int)((float)photoHeight/(float)photoWidth*config.ThumbSize);
				thumbWidth=config.ThumbSize;
			}
			else
			{
				thumbWidth=(int)((float)photoWidth/(float)photoHeight*config.ThumbSize);
				thumbHeight=config.ThumbSize;
			}
			pInfo.Thumbnail = pixbuf.ScaleSimple(thumbWidth,thumbHeight,InterpType.Bilinear);
			
			//Console.WriteLine("Leaving Sagittarius.ThumbDataFactory.GenerateThumb");
		}
		
		
		private bool DetermineType(string filename)
		{
			//Console.WriteLine("Entering Sagittarius.ThumbDataFactory.DetermineType");
			bool result;
			///////////////////////////////////////
			// TODO
			///////////////////////////////////////
			// na itt kéne valami jobb teszt a fájltípusra nézve
			string[] filenameParts = filename.Split(new char[] {'.'});
			string extension = filenameParts[filenameParts.Length-1];

			if ((extension=="dng")||(extension=="DNG"))
			{
				result = false;
			}
			else
			{
				result = true;
			}
			//Console.WriteLine("Leaving Sagittarius.ThumbDataFactory.DetermineType");
			return result;
		}
		
	public delegate void OnPhotoInfoReadyHandler (PhotoInfoFactory o, PhotoInfoReadyEventArgs args);
	public event OnPhotoInfoReadyHandler OnPhotoInfoReady;
		
	}
}
				